# mypy: disable-error-code="attr-defined"

import logging
import socket

from .error import Q3RconLibLoginError
from .packet import Request, ResponseFragment

logger = logging.getLogger(__name__)


class Base:
    def __init__(self, **kwargs):
        self.logger = logger.getChild(self.__class__.__name__)
        defaultkwargs = {
            'host': 'localhost',
            'port': 28960,
            'password': '',
            'default_timeout': 0.05,
            'timeouts': {},
        }
        kwargs = defaultkwargs | kwargs
        for attr, val in kwargs.items():
            setattr(self, attr, val)

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        Request.set_password(self.password)
        self._request = Request()

        try:
            self._host_ip = socket.gethostbyname(self.host)
        except socket.gaierror:
            raise Q3RconLibLoginError(f"unable to resolve host '{self.host}'") from None

    def send(self, cmd: str) -> str:
        """send a rcon command to the server

        Args:
            cmd (str): command string that takes the form "<cmd> <args>"

        Returns:
            str: the entire command response (all fragments) as one string
        """

        self._sock.sendto(
            self._request.to_bytes(cmd),
            (self._host_ip, self.port),
        )
        self.logger.debug(f'sending: {cmd}')

        timeout: float = self.timeouts.get(
            cmd.split(' ', maxsplit=1)[0], self.default_timeout
        )
        self._sock.settimeout(timeout)

        fragments: list[str] = []
        try:
            while resp := self._sock.recv(4096):
                try:
                    if not resp.startswith(ResponseFragment.HEADER):
                        self.logger.debug(
                            'received invalid response fragment, ignoring'
                        )
                        continue

                    resp_str = resp.removeprefix(ResponseFragment.HEADER).decode()
                    fragments.append(resp_str)
                except (ValueError, UnicodeDecodeError):
                    self.logger.debug('received invalid response fragment, ignoring')
        except TimeoutError:
            self.logger.debug('finished collecting response fragments')

        return ''.join(fragments)

    def close(self):
        self._sock.close()



"""Packet classes for handling Quake3 RCON protocol communication."""

MAGIC = b'\xff\xff\xff\xff'
RESPONSE_HEADER = MAGIC + b'print\n'


class Request:
    """RCON request packet for sending commands to the server."""

    _password: str = ''
    _password_bytes: bytes = b''

    _rcon_prefix: bytes = b'rcon '

    __slots__ = ()

    def to_bytes(self, command: str) -> bytes:
        """Convert the request to bytes for transmission.

        Args:
            command: Command to send to the server

        Returns:
            Byte representation of the RCON request
        """
        if not self._password_bytes:
            raise ValueError('Password not set. Use Request.set_password() first.')

        return (
            MAGIC
            + self._rcon_prefix
            + self._password_bytes
            + b' '
            + command.encode()
            + b'\n'
        )

    @classmethod
    def set_password(cls, password: str) -> None:
        """Set the RCON password for all Request instances.

        Args:
            password: RCON password for authentication
        """
        cls._password = password
        cls._password_bytes = password.encode()


class ResponseFragment:
    """Response packet constants and utilities."""

    HEADER = RESPONSE_HEADER

    __slots__ = ()



import functools
import time

from .error import Q3RconLibLoginError


def timeout(func):
    """Attempts a rcon login until time elapsed is greater than {Q3Rcon}._login_timeout"""

    @functools.wraps(func)
    def wrapper(*args):
        rcon, *_ = args

        err = None
        start = time.time()
        while time.time() < start + rcon.login_timeout:
            try:
                if not (resp := func(*args)):
                    raise TimeoutError('timeout attempting to login')

                if resp in (
                    'Bad rcon',
                    'Bad rconpassword.',
                    'Invalid password.',
                ):
                    raise Q3RconLibLoginError('Invalid rcon password.')
                elif (
                    resp == 'No rconpassword set on server or '
                    'password is shorter than 8 characters.\n'
                ):
                    raise Q3RconLibLoginError(
                        'No rcon password set on server or password is shorter than 8 characters.'
                    )

                err = None
                break
            except TimeoutError as e:
                err = e
                rcon.logger.debug(f'{type(e).__name__}: {e}... retrying login attempt')
                continue

        if err:
            raise Q3RconLibLoginError(str(err))

    return wrapper



# mypy: disable-error-code="attr-defined"

import dataclasses
import logging
from typing import Optional, Protocol, Self, runtime_checkable

from . import util
from .base import Base
from .error import Q3RconLibLoginError

logger = logging.getLogger(__name__)


@runtime_checkable
@dataclasses.dataclass
class DataclassProtocol(Protocol):
    pass


class Q3Rcon:
    def __init__(self, timeouts: Optional[DataclassProtocol] = None, **kwargs):
        self.logger = logger.getChild(self.__class__.__name__)
        self.login_timeout = kwargs.pop('login_timeout', 2)

        if timeouts is not None:
            if isinstance(timeouts, DataclassProtocol):
                kwargs['timeouts'] = dataclasses.asdict(timeouts)
            else:
                raise TypeError('timeouts must be an instance of a dataclass')
        self._base = Base(**kwargs)

    def __enter__(self) -> Self:
        try:
            self._login()
            self.logger.info(f'Successfully established {self}')
        except Q3RconLibLoginError as e:
            self.logger.error(f'{type(e).__name__}: {e}')
            raise
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._base.close()

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + "(hostname='{host}', port={port}, password='{password}', default_timeout={default_timeout})".format(
                **self._base.__dict__
            )
        )

    def __str__(self) -> str:
        return 'rcon connection to {host}:{port}'.format(**self._base.__dict__)

    @property
    def host(self) -> str:
        return self._base.host

    @property
    def port(self) -> int:
        return self._base.port

    @util.timeout
    def _login(self):
        return self.send('login')

    def send(self, cmd: str) -> str:
        return self._base.send(cmd)


def request_q3rcon_obj(**kwargs) -> Q3Rcon:
    """Interface entry point, use it to request a Q3Rcon object.

    Returns:
        Q3Rcon: A class implementing the Q3 Rcon protocol
    """
    cls = Q3Rcon
    return cls(**kwargs)
